/*
** Module   : ArrayAssigner.java
** Abstract : automates the assignment of each element of a BDT array
**
** Copyright (c) 2005-2009, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20051104   @23229 Provides a simple helper to automate the
**                           assignment of each element of a BaseDataType
**                           array based on an array input.
** 002 GES 20060410   @25444 All elements can be set to the same scalar
**                           value (new signature).
** 003 GES 20060412   @25486 Added a setUnknown() method.
** 004 GES 20060425   @25707 Added a new assign() signature.
** 005 ECF 20060428   @25851 Added safety code. If target array element is
**                           null, duplicate corresponding source to
**                           initialize it.
** 006 GES 20080911   @39793 Added signatures to allow primitives to be
**                           assigned from.
** 007 GES 20090625   @42956 Fix and helpers to avoid ArrayIndexOutOfBounds.
**                           This implements a new delegation model to hide
**                           all array variable dereferencing (for assignment
**                           or simple access). The helpers can thus translate
**                           or avoid Java level AIOOB exceptions and behave
**                           the way that the 4GL requires.
** 008 CA  20130209          Added support for the EXTENT statement and output EXTENT parameters.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;

/**
 * Provides a simple helper to automate the assignment of each element of a 
 * {@link BaseDataType} array based on 2 array inputs.
 */
public class ArrayAssigner
{
   /**
    * Assigns (mutates the content of) each matching element of a target
    * array based on the corresponding element of the source array.  The
    * source and target arrays must be assignable to each other (must be
    * assignment-compatible). The copy operation will stop as soon as the end
    * of either array is encountered. If either array is longer than the other,
    * then the copy will only occur for the overlapping portion. The copy
    * always starts at element 0 and finishes at the lesser of source length or
    * target length.
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The array from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, BaseDataType[] source) 
   throws ErrorConditionException,
          NullPointerException
   {
      for (int i = 0; (i < target.length && i < source.length); i++)
      {
         if (target[i] == null)
         {
            target[i] = source[i].duplicate();
         }
         else
         {
            (target[i]).assign(source[i]);
         }
      }
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, int source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new integer(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, double source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new decimal(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, String source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new character(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, boolean source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new logical(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, BaseDataType source) 
   throws ErrorConditionException,
          NullPointerException
   {
      for (int i = 0; i < target.length; i++)
      {
         if (target[i] == null)
         {
            target[i] = source.duplicate();
         }
         else
         {
            (target[i]).assign(source);
         }
      }
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target,
                                  int            idx,
                                  int            source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new integer(source));
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target,
                                  int            idx,
                                  double         source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new decimal(source));
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target,
                                  int            idx,
                                  String         source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new character(source));
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target,
                                  int            idx,
                                  boolean        source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new logical(source));
   }
   
   /**           
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target,
                                  int            idx,
                                  BaseDataType   source) 
   throws ErrorConditionException,
          NullPointerException
   {
      for (int i = idx; i < target.length; i++)
      {
         if (target[i] == null)
         {
            target[i] = source.duplicate();
         }
         else
         {
            (target[i]).assign(source);
         }
      }
   }   
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target,
                                   double         index,
                                   int            source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new integer(index), new integer(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target,
                                   double         index,
                                   double         source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new integer(index), new decimal(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target,
                                   double         index,
                                   boolean        source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new integer(index), new logical(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target,
                                   double         index,
                                   String         source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new integer(index), new character(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target,
                                   double         index,
                                   BaseDataType   source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new integer(index), source);
   }
      
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target,
                                   NumberType     index,
                                   int            source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new integer(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target,
                                   NumberType     index,
                                   double         source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new decimal(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target,
                                   NumberType     index,
                                   boolean        source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new logical(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target,
                                   NumberType     index,
                                   String         source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new character(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target,
                                   NumberType     index,
                                   BaseDataType   source) 
   throws ErrorConditionException,
          NullPointerException
   {
      int idx = -1;
      
      try
      {
         idx = subscriptHelper(target.length, index);
      }
      
      catch (ErrorConditionException ece)
      {
         // eat the first error, we will rethrow shortly
      }
      
      if (idx == -1 || ErrorManager.isPendingError())
      {
         ErrorManager.recordOrThrowError(142, "Unable to update  Field");
         
         // silently return
         return;
      }
      
      if (target[idx] == null)
      {
         target[idx] = source.duplicate();
      }
      else
      {
         (target[idx]).assign(source);
      }
   }
   
   /**
    * Dereference the given array using the given 1-based index.
    *
    * @param    array
    *           The array to dereference.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    */
   public static <B extends BaseDataType> B subscript(B[]    array,
                                                      double index)
   {
      int idx = subscriptHelper(array.length, (int) index);
      return subscriptWorker(array, idx);
   }
   
   /**
    * Dereference the given array using the given 1-based index.
    *
    * @param    array
    *           The array to dereference.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    */
   public static <B extends BaseDataType> B subscript(B[]        array,
                                                      NumberType index)
   {
      int idx = subscriptHelper(array.length, index);
      return subscriptWorker(array, idx);
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the unknown value.
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingleUnknown(BaseDataType[] target,
                                          double         index) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingleUnknown(target, new integer(index));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the unknown value.
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingleUnknown(BaseDataType[] target,
                                          NumberType     index) 
   throws ErrorConditionException,
          NullPointerException
   {
      int idx = -1;
      
      try
      {
         idx = subscriptHelper(target.length, index);
      }
      
      catch (ErrorConditionException ece)
      {
         // eat the first error, we will rethrow shortly
      }
      
      if (idx == -1 || ErrorManager.isPendingError())
      {
         ErrorManager.recordOrThrowError(142, "Unable to update  Field");
         
         // silently return
         return;
      }
      
      target[idx].setUnknown();
   }
   
   /**                     
    * Sets all elements of a target array to the <code>unknown value</code>.
    *
    * @param    target
    *           The array to be modified. Must not be <code>null</code>.
    */
   public static void assignUnknown(BaseDataType[] target) 
   {
      for (int i = 0; i < target.length; i++)
      {
         target[i].setUnknown();
      }
   }
   
   /**
    * Dereference the given array using the given 0-based index.
    *
    * @param    array
    *           The array to dereference.
    * @param    idx
    *           The 0-based index into the array. No bounds checking is done.
    */
   private static <B extends BaseDataType> B subscriptWorker(B[] array,
                                                             int idx)
   {
      if (idx == -1 || ErrorManager.isPendingError())
      {
         Class<?> cls = array.getClass().getComponentType();
         
         // silently return unknown value (safest thing)
         return (B) BaseDataType.generateUnknown(cls);
      }
      
      return array[idx];
   }
   
   /**
    * Generate a valid 0-based integral subscript from the 1-based numeric 
    * value given. Bounds checking is done based on the target array length.
    *
    * @param    len
    *           The length of the target array that will be dereferenced. This
    *           will be used to detect the array size and implement the proper
    *           bounds checking.
    * @param    idx
    *           The 1-based index to be checked and modified.
    *
    * @return   The 0-based integral subscript or -1 if the index is the
    *           <code>unknown value</code> or if the index is out of bounds
    *           in silent error mode.
    *
    * @throws   ErrorConditionException
    *           If the index is out of bounds.
    */
   private static int subscriptHelper(int len, NumberType idx)
   throws ErrorConditionException
   {
      if (idx.isUnknown())
      {
         return -1;
      }
      
      return subscriptHelper(len, idx.intValue());
   }
   
   /**
    * Generate a valid 0-based integral subscript from the 1-based numeric 
    * value given. Bounds checking is done based on the target array length.
    *
    * @param    len
    *           The length of the target array that will be dereferenced. This
    *           will be used to detect the array size and implement the proper
    *           bounds checking.
    * @param    idx
    *           The 1-based index to be checked and modified.
    *
    * @return   The 0-based integral subscript or -1 if the index is out
    *           of bounds in silent error mode.
    *
    * @throws   ErrorConditionException
    *           If the index is out of bounds.
    */
   private static int subscriptHelper(int len, int idx)
   throws ErrorConditionException
   {
      int sub = -1;
      
      if (idx > 0)
      {
        sub = idx - 1;
      }
      
      if (sub < 0 || sub >= len)
      {
         // error out!
         String spec   = "Array subscript %d is out of range";
         String errmsg = String.format(spec, sub + 1);
         
         ErrorManager.recordOrThrowError(26, errmsg);
      }
      
      return sub;
   }

   /**
    * Implement the EXTENT statement. This resizes the given array to the new size and returns
    * the new instance.
    * <p>
    * If the given array is not of size zero, then an error is raised.  Else, all the elements
    * of the array will be set to unknown.
    * 
    * @param    array
    *           The dynamic-extent array to be resized.
    * @param    newSize
    *           The array's new size.
    * 
    * @return   The resized array.
    */
   public static <T extends BaseDataType> T[] resize(T[] array, int newSize)
   {
      return resize(array, new integer(newSize));
   }

   /**
    * Implement the EXTENT statement. This resizes the given array to the new size and returns
    * the new instance.
    * <p>
    * If the given array is not of size zero (and the newSize is not unknown), then an error is 
    * raised.  Else, the array will be resized and all the elements will be set to the default
    * value.
    * <p>
    * The new extent value must be from the [1..28000] interval.
    * 
    * @param    array
    *           The dynamic-extent array to be resized. Must not be null. If the array was not
    *           resized yet, must be 0-length.
    * @param    newSize
    *           The array's new size.
    * 
    * @return   The resized array.
    */
   public static <T extends BaseDataType> T[] resize(T[] array, integer newSize)
   {
      int size = newSize.isUnknown() ? 0 : newSize.intValue();
      if (!newSize.isUnknown())
      {
         if (array.length != 0)
         {
            final String msg = "Indeterminate extent is already fixed to a dimension of %d"; 
            ErrorManager.recordOrThrowError(13738, String.format(msg, array.length), false);
            return array;
         }

         if (size <= 0 || size > 28000)
         {
            final String msg = "Indeterminate extent dimension %d is out of range 1 to 28000"; 
            ErrorManager.recordOrThrowError(11385, String.format(msg, size), false);
            return array;
         }
      }
      
      Class<? extends BaseDataType> cls = 
         (Class<? extends BaseDataType>) array.getClass().getComponentType();
      T[] copy = (T[]) Array.newInstance(cls, size);
      for (int i = 0; i < copy.length; i++)
      {
         copy[i] = (T) BaseDataType.generateDefault(cls);
      }
      
      return copy;
   }

   /**
    * Get a copy of the given array.
    * 
    * @param    array
    *           The array for which a copy is needed.
    * 
    * @return   The copy.
    */
   public static <T extends BaseDataType> T[] copyOf(T[] array)
   {
      Class<? extends BaseDataType> cls = 
               (Class<? extends BaseDataType>) array.getClass().getComponentType();

      T[] copy = (T[]) Array.newInstance(cls, array.length);
      for (int i = 0; i < copy.length; i++)
      {
         copy[i] = (T) BaseDataType.generateUnknown(cls);
         copy[i].assign(array[i]);
      }
      
      return copy;
   }
   
   /**
    * Get the length of the given dynamic-extent variable.
    * 
    * @param    array
    *           The dynamic-extent variable.
    *           
    * @return   If the array's length is zero, return unknown. Else, the actual size.
    */
   public static integer lengthOf(BaseDataType[] array)
   {
      return new integer(array.length == 0 ? null : array.length);
   }
}
